// r8c-writer is a R8C microcontroller serial writer program.
//
// Copyright (C) 2016 T-semi
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#define _POSIX_C_SOURCE 199309L
#ifdef R8C_WRITER_LONGOPT
#  define _GNU_SOURCE
#  include <getopt.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdint.h>
#include <string.h>
#include <assert.h>

#ifdef _WIN32
#  include <windows.h>
#else
#  include <unistd.h>
#  include <time.h>
#  include <termios.h>
#  include <fcntl.h>
#  include <poll.h>
#  include <sys/types.h>
#  include <sys/stat.h>
#  include <sys/ioctl.h>
#endif

#define WRSTAT_CONNOK   0x00000001
#define WRSTAT_CONNFAIL 0x00000002
#define WRSTAT_IDSTAT   0x0000000c
#define WRSTAT_IDFAIL   0x00000004
#define WRSTAT_NOIDCK   0x00000008
#define WRSTAT_IDOK     0x0000000c
#define WRSTAT_COMERROR 0x00000010

#define WRPIN_RTS       0
#define WRPIN_DTR       1
#define WRPIN_INVERT    2

#define WRSREG_READY    0x00000080
#define WRSREG_ERER     0x00000020
#define WRSREG_WRER     0x00000010
#define WRSREG_IDSTAT   0x00000c00
#define WRSREG_IDFAIL   0x00000400
#define WRSREG_IDOK     0x00000c00
#define WRSREG_READERR  0xffffffff

#define WRID_UNDEFINED  0xffffffffffffffff

#define WRFLAG_NOECHO   0x00000001

#ifdef R8C_WRITER_CONFIG_AT
char *path_to_conf[] = {R8C_WRITER_CONFIG_AT};
#endif

char *myname = "r8c-writer";
char *whoami = "R8C microcontroller writer";
char *version = "0.2.1";
char *copyright = "Copyright (C) 2016 T-semi";
char *lisence = "This free software comes with ABSOLUTELY NO WARRANTY.\n"
                "See the source for details.";

struct myoption{
    char shortname;
    char *longname;
    int ifparam;
    char *description;
    char *param;
};

struct myoption arg_def[] = {
#define OPT_HELP 0
    {'h', "help", 0,
        "Show this help, and exit", NULL},
#define OPT_VERSION 1
    {'V', "version", 0,
        "Show version information and copyright, and exit", NULL},
#define OPT_SHOWINFO 2
    {'I', "info", 0,
        "Show config information of target and/or writer and exit", NULL},
#define OPT_TEST 3
    {'T', "test", 0,
        "Get firmware version and exit", NULL},
#define OPT_EXEC 4
    {'E', "exec", 0,
        "Reset target and execute program", NULL},
#define OPT_WRITER 5
    {'w', "writer", 1,
        "Give writing cable's name", NULL},
#define OPT_TARGET 6
    {'t', "target", 1,
        "Give target's name", NULL},
#define OPT_CONFIG 7
    {'c', "config", 1,
        "Configuration file name", NULL},
#define OPT_DEVICE 8
    {'d', "device", 1,
        "Specify a device file. Otherwise, default device is used.", NULL},
#define OPT_UNLOCKID 9
    {'i', "id", 1,
        "Specify a id for id ckeck features."
        "Otherwise, 0xffffffffffffff is used.", NULL}
};

char *filename = NULL;

struct writer_context{
#ifdef _WIN32
    HANDLE fh;
#else
    int fd;
#endif
    uint64_t id;
    int status;
    int pin_reset;
    int pin_mode;
    int flags;
    int timeout;
};

struct chaind_page{
    uint16_t addr;
    char data[256];
    struct chaind_page *next;
};

struct writer_desc_chain{
    char *name;
    char *default_path;
    struct writer_context config;
    struct writer_desc_chain *next;
};

struct flash_block{
    uint32_t start;
    uint32_t size;
};

struct target_mmap_chain{
    char *friendry_name;
    int namen;
    char **names;
    int blockn;
    struct flash_block *blocks;
    struct target_mmap_chain *next;
};

struct vl_string{
    size_t bufsize;
    size_t length;
    char *buf;
};

void vl_string_new(struct vl_string *s, size_t size){
    s->bufsize = size;
    s->length = 0;
    s->buf = malloc(size);
    if(!s->buf){
        fprintf(stderr, "malloc: error\n");
        exit(1);
    }
}

void vl_string_free(struct vl_string *s){
    free(s->buf);
    s->buf = NULL;
    s->bufsize = 0;
    s->length = 1;
}

void vl_string_push(struct vl_string *s, char c){
    assert(s->bufsize >= s->length);
    s->length++;
    if(s->length > s->bufsize){
        s->bufsize *= 2;
        char *nb = realloc(s->buf, s->bufsize);
        if(!nb){
            fprintf(stderr, "realloc: error\n");
            exit(1);
        }
        s->buf = nb;
    }
    s->buf[s->length-1] = c;
}

struct writer_desc_chain *writers = NULL;
struct target_mmap_chain *targets = NULL;

void my_msleep(unsigned int ms){
#ifdef _WIN32
    Sleep(ms);
#else
    struct timespec ts;
    ts.tv_sec = ms / 1000;
    ts.tv_nsec = ms % 1000 * 1000000;

    if(nanosleep(&ts, NULL) < 0){
        perror("nanosleep");
        exit(1);
    }
#endif
}

void strlower(char *str){
    while(*str){
        if('A' <= *str && *str <= 'Z'){
            *str += 'a' - 'A';
        }
        str++;
    }
}

void strctarget(char *str){
    strlower(str);
    while(*str){
        if(!('0' <= *str && *str <= '9') && !('a' <= *str && *str <= 'z')){
            char *s2 = str;
            while(*s2){
                s2[0] = s2[1];
                s2++;
            }
        }else{
            str++;
        }
    }
}

int writer_if_ready(struct writer_context *wc){
    return wc->status & WRSTAT_CONNOK &&
           !(wc->status & WRSTAT_COMERROR) &&
           !(wc->status & WRSTAT_CONNFAIL);
}

int writer_unlocked(struct writer_context *wc){
    int ids = wc->status & WRSTAT_IDSTAT;
    return ids == WRSTAT_NOIDCK || ids == WRSTAT_IDOK;
}

void set_reset(struct writer_context *wc, int level){
#ifdef _WIN32
    DWORD op;
    if(wc->pin_reset & WRPIN_DTR){
        if(!(wc->pin_reset & WRPIN_INVERT) == !level){
            op = SETDTR;
        }else{
            op = CLRDTR;
        }
    }else{
        if(!(wc->pin_reset & WRPIN_INVERT) == !level){
            op = SETRTS;
        }else{
            op = CLRRTS;
        }
    }
    if(!EscapeCommFunction(wc->fh, op)){
        fprintf(stderr, "reset pin driving error");
        exit(1);
    }
#else
    int mb = (wc->pin_reset & WRPIN_DTR) ? TIOCM_DTR : TIOCM_RTS;
    int rv = ioctl(wc->fd,
            (!(wc->pin_reset&WRPIN_INVERT)==!level)?TIOCMBIS:TIOCMBIC, &mb);
    if(rv < 0){
        perror("ioctl");
        exit(1);
    }
#endif
}

void set_mode(struct writer_context *wc, int level){
#ifdef _WIN32
    DWORD op;
    if(wc->pin_mode & WRPIN_DTR){
        if(!(wc->pin_mode & WRPIN_INVERT) == !level){
            op = SETDTR;
        }else{
            op = CLRDTR;
        }
    }else{
        if(!(wc->pin_mode & WRPIN_INVERT) == !level){
            op = SETRTS;
        }else{
            op = CLRRTS;
        }
    }
    if(!EscapeCommFunction(wc->fh, op)){
        fprintf(stderr, "mode pin driving error");
        exit(1);
    }
#else
    int mb = (wc->pin_mode & WRPIN_DTR) ? TIOCM_DTR : TIOCM_RTS;
    int rv = ioctl(wc->fd,
            (!(wc->pin_mode&WRPIN_INVERT)==!level)?TIOCMBIS:TIOCMBIC, &mb);
    if(rv < 0){
        perror("ioctl");
        exit(1);
    }
#endif
}

int send_data(struct writer_context *wc, size_t datlen, char *dat){
    size_t left = datlen;
    char *dtop = dat;
    char buf[1024];

    if(!writer_if_ready(wc)){
        return -1;
    }

#ifdef _WIN32
    if(!SetCommMask(wc->fh, EV_TXEMPTY)){
        fprintf(stderr, "SetCommMask error\n");
        exit(1);
    }
#endif

    while(left){
#ifdef _WIN32
        DWORD written;
        OVERLAPPED ols = {0};
        ols.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
        if(!ols.hEvent){
            fprintf(stderr, "CreateEvent error\n");
            exit(1);
        }
        if(!WriteFile(wc->fh, dtop, left, &written, &ols)){
            if(GetLastError() == ERROR_IO_PENDING){
                if(!GetOverlappedResult(wc->fh, &ols, &written, TRUE)){
                    fprintf(stderr, "WriteFile error"
                                    "while GetOverlappedResult\n");
                    exit(1);
                }
            }else{
                fprintf(stderr, "WriteFile error\n");
                exit(1);
            }
        }
        if(!CloseHandle(ols.hEvent)){
            fprintf(stderr, "CloseHandle error\n");
            exit(1);
        }
#else
        ssize_t written = write(wc->fd, dtop, left);
        if(written < 0){
            perror("write");
            exit(1);
        }
#endif
        left -= written;
        dtop += written;
    }

#ifdef _WIN32
    DWORD evmask;
    OVERLAPPED ols = {0};
    ols.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
    if(!ols.hEvent){
        fprintf(stderr, "CreateEvent error\n");
        exit(1);
    }
    if(!WaitCommEvent(wc->fh, &evmask, &ols)){
        if(GetLastError() == ERROR_IO_PENDING){
            DWORD t;
            if(!GetOverlappedResult(wc->fh, &ols, &t, TRUE)){
                fprintf(stderr, "WaitCommEvent error"
                                "while GetOverlappedResult\n");
                exit(1);
            }
        }else{
            fprintf(stderr, "WaitCommEvent error\n");
            exit(1);
        }
    }
    if(!CloseHandle(ols.hEvent)){
        fprintf(stderr, "CloseHandle error\n");
        exit(1);
    }
#else
    if(tcdrain(wc->fd)){
        perror("tcdrain");
        exit(1);
    }
#endif

    if(wc->flags & WRFLAG_NOECHO){
        return 0;
    }

    left = datlen;
    while(left){
        size_t nread = left;
        if(nread > sizeof(buf)){
            nread = sizeof(buf);
        }

#ifdef _WIN32
        DWORD read_n;
        OVERLAPPED ols = {0};
        ols.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
        if(!ols.hEvent){
            fprintf(stderr, "CreateEvent error\n");
            exit(1);
        }
        if(!ReadFile(wc->fh, buf, nread, &read_n, &ols)){
            if(GetLastError() == ERROR_IO_PENDING){
                DWORD rv = WaitForSingleObject(ols.hEvent, wc->timeout);
                if(rv == WAIT_OBJECT_0){
                    if(!GetOverlappedResult(wc->fh, &ols, &read_n, FALSE)){
                        fprintf(stderr, "ReadFile error"
                                        "while GetOverlappedResult\n");
                        exit(1);
                    }
                }else if(rv == WAIT_TIMEOUT){
                    if(!CloseHandle(ols.hEvent)){
                        fprintf(stderr, "CloseHandle error\n");
                        exit(1);
                    }
                    wc->status |= WRSTAT_CONNFAIL;
                    return 1;
                }else{
                    fprintf(stderr, "ReadFile error"
                                    "while WaitForSingleObject\n");
                    exit(1);
                }
            }else{
                fprintf(stderr, "ReadFile error\n");
                exit(1);
            }
        }
        if(!CloseHandle(ols.hEvent)){
            fprintf(stderr, "CloseHandle error\n");
            exit(1);
        }
#else
        struct pollfd pfd;
        pfd.fd = wc->fd;
        pfd.events = POLLIN;
        if(poll(&pfd, 1, wc->timeout) < 0){
            perror("poll");
            exit(1);
        }
        if(!pfd.revents){
            wc->status |= WRSTAT_CONNFAIL;
            return 1;   //timeout
        }else if(pfd.revents != POLLIN){
            fprintf(stderr, "poll error\n");
            exit(1);
        }

        ssize_t read_n = read(wc->fd, buf, nread);
        if(read_n < 0){
            perror("read");
            exit(1);
        }
#endif
        left -= read_n;
    }

    return 0;
}

int recv_data(struct writer_context *wc, size_t req, char *buf){
    size_t left = req;
    char *btop = buf;

    if(!writer_if_ready(wc)){
        return -1;
    }

    while(left){
#ifdef _WIN32
        DWORD read_n;
        OVERLAPPED ols = {0};
        ols.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
        if(!ols.hEvent){
            fprintf(stderr, "CreateEvent error\n");
            exit(1);
        }
        if(!ReadFile(wc->fh, btop, left, &read_n, &ols)){
            if(GetLastError() == ERROR_IO_PENDING){
                DWORD rv = WaitForSingleObject(ols.hEvent, wc->timeout);
                if(rv == WAIT_OBJECT_0){
                    if(!GetOverlappedResult(wc->fh, &ols, &read_n, FALSE)){
                        fprintf(stderr, "ReadFile error"
                                        "while GetOverlappedResult\n");
                        exit(1);
                    }
                }else if(rv == WAIT_TIMEOUT){
                    if(!CloseHandle(ols.hEvent)){
                        fprintf(stderr, "CloseHandle error\n");
                        exit(1);
                    }
                    wc->status |= WRSTAT_COMERROR;
                    return 1;
                }else{
                    fprintf(stderr, "ReadFile error"
                                    "while WaitForSingleObject\n");
                    exit(1);
                }
            }else{
                fprintf(stderr, "ReadFile error\n");
                exit(1);
            }
        }
        if(!CloseHandle(ols.hEvent)){
            fprintf(stderr, "CloseHandle error\n");
            exit(1);
        }
#else
        struct pollfd pfd;
        pfd.fd = wc->fd;
        pfd.events = POLLIN;
        if(poll(&pfd, 1, wc->timeout) < 0){
            perror("poll");
            exit(1);
        }
        if(!pfd.revents){
            wc->status |= WRSTAT_COMERROR;
            return 1;   //timeout
        }else if(pfd.revents != POLLIN){
            fprintf(stderr, "poll error\n");
            exit(1);
        }

        ssize_t read_n = read(wc->fd, btop, left);
        if(read_n < 0){
            perror("read");
            exit(1);
        }
#endif
        left -= read_n;
        btop += read_n;
    }

    my_msleep(2);       //long stop bits
    return 0;
}

int program_read_page(struct writer_context *wc, uint16_t addr, char *buf){
    char cmd = 0xff;

    if(!writer_if_ready(wc) || !writer_unlocked(wc)){
        return -1;
    }

    if(send_data(wc, 1, &cmd)){
        return 1;
    }
    if(send_data(wc, 2, (char *)&addr)){
        return 1;
    }
    if(recv_data(wc, 256, buf)){
        return 2;
    }
    return 0;
}

int program_read_stat(struct writer_context *wc){
    char cmd = 0x70;
    int rbuf = 0;

    if(!writer_if_ready(wc)){
        return WRSREG_READERR;
    }

    if(send_data(wc, 1, &cmd)){
        return WRSREG_READERR;
    }
    if(recv_data(wc, 2, (char *)&rbuf)){
        return WRSREG_READERR;
    }

    return rbuf;
}

int program_clear_stat(struct writer_context *wc){
    char cmd = 0x50;

    if(!writer_if_ready(wc)){
        return -1;
    }

    if(send_data(wc, 1, &cmd)){
        return 1;
    }

    int rv = program_read_stat(wc);
    if(rv == WRSREG_READERR || (rv & WRSREG_ERER) || (rv & WRSREG_WRER)){
        return 3;
    }else{
        return 0;
    }
}

int program_write_page(struct writer_context *wc, uint16_t addr, char *buf){
    char cmd = 0x41;
    int rv;

    if(!writer_if_ready(wc) || !writer_unlocked(wc)){
        return -1;
    }

    rv = program_read_stat(wc);
    if(rv == WRSREG_READERR){
        return 2;
    }
    while(!(rv & WRSREG_READY)){
        rv = program_read_stat(wc);
        if(rv == WRSREG_READERR){
            return 2;
        }
    }

    rv = program_clear_stat(wc);
    if(rv){
        if(rv == 1 || rv == 3){
            return rv;
        }else{
            return 4;
        }
    }

    if(send_data(wc, 1, &cmd)){
        return 10;
    }
    if(send_data(wc, 2, (char *)&addr)){
        return 10;
    }
    if(send_data(wc, 256, buf)){
        return 11;
    }

    rv = program_read_stat(wc);
    if(rv == WRSREG_READERR){
        return 22;
    }
    while(!(rv & WRSREG_READY)){
        rv = program_read_stat(wc);
        if(rv == WRSREG_READERR){
            return 22;
        }
    }
    if(rv == WRSREG_READERR){
        return 20;
    }else if(rv & WRSREG_WRER){
        return 21;
    }

    return 0;
}

int program_erase_block(struct writer_context *wc, uint16_t addr){
    char cmd = 0x20;
    char stp = 0xd0;
    int rv;

    if(!writer_if_ready(wc) || !writer_unlocked(wc)){
        return -1;
    }

    rv = program_read_stat(wc);
    if(rv == WRSREG_READERR){
        return 2;
    }
    while(!(rv & WRSREG_READY)){
        rv = program_read_stat(wc);
        if(rv == WRSREG_READERR){
            return 2;
        }
    }

    rv = program_clear_stat(wc);
    if(rv){
        if(rv == 1 || rv == 3){
            return rv;
        }else{
            return 4;
        }
    }

    if(send_data(wc, 1, &cmd)){
        return 10;
    }
    if(send_data(wc, 2, (char *)&addr)){
        return 10;
    }
    if(send_data(wc, 1, &stp)){
        return 11;
    }

    rv = program_read_stat(wc);
    if(rv == WRSREG_READERR){
        return 22;
    }
    while(!(rv & WRSREG_READY)){
        rv = program_read_stat(wc);
        if(rv == WRSREG_READERR){
            return 22;
        }
    }
    if(rv == WRSREG_READERR){
        return 20;
    }else if(rv & WRSREG_ERER){
        return 21;
    }

    return 0;
}

int program_erase_all(struct writer_context *wc){
    char cmd[2] = {0xa7, 0xd0};
    int rv;

    if(!writer_if_ready(wc) || !writer_unlocked(wc)){
        return -1;
    }

    rv = program_read_stat(wc);
    if(rv == WRSREG_READERR){
        return 2;
    }
    while(!(rv & WRSREG_READY)){
        rv = program_read_stat(wc);
        if(rv == WRSREG_READERR){
            return 2;
        }
    }

    rv = program_clear_stat(wc);
    if(rv){
        if(rv == 1 || rv == 3){
            return rv;
        }else{
            return 4;
        }
    }

    if(send_data(wc, 2, cmd)){
        return 11;
    }

    rv = program_read_stat(wc);
    if(rv == WRSREG_READERR){
        return 22;
    }
    while(!(rv & WRSREG_READY)){
        rv = program_read_stat(wc);
        if(rv == WRSREG_READERR){
            return 22;
        }
    }
    if(rv == WRSREG_READERR){
        return 20;
    }else if(rv & WRSREG_ERER){
        return 21;
    }

    return 0;
}

int program_check_id(struct writer_context *wc, uint64_t id){
    char cmd[] = {0xf5, 0xdf, 0xff, 0x00, 0x07};

    if(!writer_if_ready(wc)){
        return -1;
    }

    if(send_data(wc, 5, cmd)){
        return 1;
    }
    if(send_data(wc, 7, (char *)&id)){
        return 1;
    }

    int sr = program_read_stat(wc);
    if(sr == WRSREG_READERR){
        return 3;
    }else if((sr & WRSREG_IDSTAT) != WRSREG_IDOK){
        wc->status &= ~WRSTAT_IDSTAT;
        wc->status |= WRSTAT_IDFAIL;
        return 4;
    }else{
        wc->status &= ~WRSTAT_IDSTAT;
        wc->status |= WRSTAT_IDOK;
        return 0;
    }
}

//buf size >= 9
int program_get_version(struct writer_context *wc, char *buf){
    char cmd = 0xfb;

    if(!writer_if_ready(wc)){
        return -1;
    }

    buf[8] = 0;
    if(send_data(wc, 1, &cmd)){
        return 1;       //hardlink error
    }
    if(recv_data(wc, 8, buf)){
        return 2;       //no response
    }
    if(buf[0] != 'V' || buf[1] != 'E' || buf[2] != 'R' ||
       buf[3] != '.' || buf[5] != '.'){
        return 3;       //application error
    }
    if(buf[4] < '0' || buf[4] > '9' || buf[6] < '0' || buf[6] > '9' ||
       buf[7] < '0' || buf[7] > '9'){
        return 3;       //application error
    }
    return 0;
}

void program_reset(struct writer_context *wc){
    wc->status = 0;
    set_mode(wc, 1);
    set_reset(wc, 0);
    my_msleep(35);
    set_reset(wc, 1);
}

void serdev_init(struct writer_context *wc){
#ifdef _WIN32
    DCB dcb = {0};
    dcb.DCBlength = sizeof(dcb);
    if(!GetCommState(wc->fh, &dcb)){
        fprintf(stderr, "GetCommState error\n");
        exit(1);
    }
    if(dcb.DCBlength != sizeof(dcb)){
        fprintf(stderr, "DCB structure format error\n");
        exit(1);
    }
    dcb.BaudRate = CBR_9600;
    dcb.fBinary = TRUE;
    dcb.fParity = FALSE;
    dcb.fOutxCtsFlow = FALSE;
    dcb.fOutxDsrFlow = FALSE;
    if(dcb.fDtrControl != DTR_CONTROL_DISABLE &&
       dcb.fDtrControl != DTR_CONTROL_ENABLE){
        dcb.fDtrControl = DTR_CONTROL_DISABLE;
    }
    dcb.fDsrSensitivity = FALSE;
    dcb.fTXContinueOnXoff = TRUE;
    dcb.fOutX = FALSE;
    dcb.fInX = FALSE;
    dcb.fErrorChar = FALSE;
    dcb.fNull = FALSE;
    if(dcb.fRtsControl != RTS_CONTROL_DISABLE &&
       dcb.fRtsControl != RTS_CONTROL_ENABLE){
        dcb.fRtsControl = RTS_CONTROL_DISABLE;
    }
    dcb.fAbortOnError = FALSE;
    //dcb.fDummy2
    dcb.wReserved = 0;
    //dcb.XonLim
    //dcb.XoffLim
    dcb.ByteSize = 8;
    dcb.Parity = NOPARITY;
    dcb.StopBits = ONESTOPBIT;
    //dcb.XonChar
    //dcb.XoffChar
    //dcb.ErrorChar
    //dcb.EofChar
    //dcb.EvtChar
    //dcb.wReserved1
    if(!SetCommState(wc->fh, &dcb)){
        fprintf(stderr, "SetCommState error\n");
        exit(1);
    }

    COMMTIMEOUTS to;
    to.ReadIntervalTimeout = 0;
    to.ReadTotalTimeoutMultiplier = 0;
    to.ReadTotalTimeoutConstant = 0;
    to.WriteTotalTimeoutMultiplier = 0;
    to.WriteTotalTimeoutConstant = 0;
    if(!SetCommTimeouts(wc->fh, &to)){
        fprintf(stderr, "SetCommTimeouts error\n");
        exit(1);
    }

    //flush recving buffer
    while(1){
        DWORD eflg;
        COMSTAT cstat;
        if(!ClearCommError(wc->fh, &eflg, &cstat)){
            fprintf(stderr, "ClearCommError error\n");
            exit(1);
        }
        if(!cstat.cbInQue){
            break;
        }

        DWORD read_n;
        OVERLAPPED ols = {0};
        char rb;
        ols.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
        if(!ols.hEvent){
            fprintf(stderr, "CreateEvent error\n");
            exit(1);
        }
        if(!ReadFile(wc->fh, &rb, 1, &read_n, &ols)){
            if(GetLastError() == ERROR_IO_PENDING){
                if(!GetOverlappedResult(wc->fh, &ols, &read_n, TRUE)){
                    fprintf(stderr, "ReadFile error"
                                    "while GetOverlappedResult\n");
                    exit(1);
                }
            }else{
                fprintf(stderr, "ReadFile error\n");
                exit(1);
            }
        }
        if(!CloseHandle(ols.hEvent)){
            fprintf(stderr, "CloseHandle error\n");
            exit(1);
        }
    }
#else
    //termio settings
    struct termios tios;
    memset(&tios, 0, sizeof(tios));
    tios.c_iflag = IGNBRK;
    tios.c_oflag = 0;
    tios.c_cflag = CS8 | CREAD | CLOCAL;
    tios.c_lflag = NOFLSH;
    if(cfsetispeed(&tios, B9600)){
        perror("cfsetispeed");
        exit(1);
    }
    if(cfsetospeed(&tios, B9600)){
        perror("cfsetospeed");
        exit(1);
    }
    tcsetattr(wc->fd, TCSANOW, &tios);      //Ignore Error
    if(tcflush(wc->fd, TCIOFLUSH)){
        perror("tcflush");
        exit(1);
    }
#endif
}

int program_init(struct writer_context *wc){
    serdev_init(wc);

    set_mode(wc, 0);
    set_reset(wc, 0);       //mode and reset pins to low

    my_msleep(35);
    set_reset(wc, 1);       //reset pin to high
    my_msleep(35);
    set_mode(wc, 1);        //mode pin to high = ready to serial com
    my_msleep(270);         //wait for entering mode 3

    //set baud rate
    char cmd;
    char res;
    wc->status = WRSTAT_CONNOK;
    cmd = 0;
    for(int i = 0; i < 16; i++){
        if(send_data(wc, 1, &cmd)){
            wc->status = (wc->status & ~WRSTAT_CONNOK) | WRSTAT_CONNFAIL;
            return 1;   //hardlink error
        }
        my_msleep(20);
    }
#ifdef _WIN32
    while(1){
        DWORD eflg;
        COMSTAT cstat;
        if(!ClearCommError(wc->fh, &eflg, &cstat)){
            fprintf(stderr, "ClearCommError error\n");
            exit(1);
        }
        if(!cstat.cbInQue){
            break;
        }

        DWORD read_n;
        OVERLAPPED ols = {0};
        char rb;
        ols.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
        if(!ols.hEvent){
            fprintf(stderr, "CreateEvent error\n");
            exit(1);
        }
        if(!ReadFile(wc->fh, &rb, 1, &read_n, &ols)){
            if(GetLastError() == ERROR_IO_PENDING){
                if(!GetOverlappedResult(wc->fh, &ols, &read_n, TRUE)){
                    fprintf(stderr, "ReadFile error"
                                    "while GetOverlappedResult\n");
                    exit(1);
                }
            }else{
                fprintf(stderr, "ReadFile error\n");
                exit(1);
            }
        }
        if(!CloseHandle(ols.hEvent)){
            fprintf(stderr, "CloseHandle error\n");
            exit(1);
        }
    }
#else
    if(tcflush(wc->fd, TCIOFLUSH)){
        perror("tcflush");
        exit(1);
    }
#endif
    cmd = 0xb0;
    if(send_data(wc, 1, &cmd)){
        wc->status = (wc->status & ~WRSTAT_CONNOK) | WRSTAT_CONNFAIL;
        return 1;       //hardlink error
    }
    if(recv_data(wc, 1, &res)){
        wc->status = (wc->status & ~WRSTAT_CONNOK) | WRSTAT_CONNFAIL;
        return 2;       //no response
    }
    if(res != cmd){
        wc->status = (wc->status & ~WRSTAT_CONNOK) | WRSTAT_CONNFAIL;
        return 3;       //application error
    }

    int st;
    st = program_read_stat(wc);
    if(st == WRSREG_READERR){
        wc->status = (wc->status & ~WRSTAT_CONNOK) | WRSTAT_CONNFAIL;
        return 3;       //application error
    }

    wc->status &= ~WRSTAT_IDSTAT;
    if((st & WRSREG_IDSTAT) == WRSREG_IDOK){
        wc->status |= WRSTAT_NOIDCK;
        wc->id = 0x00ffffffffffffff;
    }else{
        wc->id = WRID_UNDEFINED;
    }
    return 0;
}

int program_swrite_page(struct writer_context *wc, uint16_t addr, char *buf){
    char tbuf[256];
    int rv;

    if(!writer_if_ready(wc) || !writer_unlocked(wc)){
        return -1;
    }

    rv = program_write_page(wc, addr, buf);
    if(rv){
        int efbup =
           (wc->status & WRSTAT_COMERROR) |
           (wc->status & WRSTAT_CONNFAIL);
        wc->status &= ~WRSTAT_COMERROR;
        wc->status &= ~WRSTAT_CONNFAIL;
        if(rv < 10){
            if(program_read_page(wc, addr, buf)){
                wc->status |= efbup;
                return 11;
            }else{
                wc->status |= efbup;
                return 21;
            }
        }else{
            if(program_read_page(wc, addr, buf)){
                wc->status |= efbup;
                return 12;
            }else{
                wc->status |= efbup;
                return 22;
            }
        }
    }

    rv = program_read_page(wc, addr, tbuf);
    if(rv){
        return 13;
    }else{
        if(memcmp(buf, tbuf, 256)){
            memcpy(buf, tbuf, 256);
            return 23;
        }else{
            return 0;
        }
    }
}

void pagedump(char *_page){
    int i;
    unsigned char *page = (unsigned char *)_page;
    for(i=0; i<256; i+=16){
        printf("%02x %02x %02x %02x %02x %02x %02x %02x "
               "%02x %02x %02x %02x %02x %02x %02x %02x\n",
               page[i],    page[i+1],  page[i+2],  page[i+3],
               page[i+4],  page[i+5],  page[i+6],  page[i+7],
               page[i+8],  page[i+9],  page[i+10], page[i+11],
               page[i+12], page[i+13], page[i+14], page[i+15]);
    }
}

void cpagedump(struct chaind_page *cp){
    while(cp){
        printf("page[%08x]:\n", cp->addr);
        pagedump(cp->data);
        cp = cp->next;
    }
}

void free_cpage(struct chaind_page *cp){
    struct chaind_page *np;
    while(cp){
        np = cp->next;
        free(cp);
        cp = np;
    }
}

struct chaind_page *srecfp2cpage(FILE *fp){
    struct chaind_page *rp = NULL;
    char linebuf[1024];

    while(fgets(linebuf, 1024, fp)){
        if(linebuf[0] != 'S'){
            continue;
        }
        if(linebuf[1] < '1' || '3' < linebuf[1]){
            continue;
        }
        int al = linebuf[1] - '0' + 1;
        int bc = 0;
        uint8_t sum = 0;
        for(int i = 2; i < 4; i++){
            bc <<= 4;
            if('0' <= linebuf[i] && linebuf[i] <= '9'){
                bc |= linebuf[i] - '0';
            }else if('A' <= linebuf[i] && linebuf[i] <= 'F'){
                bc |= linebuf[i] - 'A' + 10;
            }else{
                free_cpage(rp);
                return NULL;
            }
        }
        sum += bc;
        if(bc <= al){
            free_cpage(rp);
            return NULL;
        }
        uint32_t addr = 0;
        for(int i = 4; i < al * 2 + 4; i++){
            addr <<= 4;
            if('0' <= linebuf[i] && linebuf[i] <= '9'){
                addr |= linebuf[i] - '0';
            }else if('A' <= linebuf[i] && linebuf[i] <= 'F'){
                addr |= linebuf[i] - 'A' + 10;
            }else{
                free_cpage(rp);
                return NULL;
            }
            if(i & 1){
                sum += addr & 0xff;
            }
        }
        if(addr > 0xffffff || addr+(bc-al) > 0x1000000){
            free_cpage(rp);
            return NULL;
        }
        struct chaind_page *tp = rp;
        if(tp == NULL || tp->addr > addr >> 8){
            tp = malloc(sizeof(struct chaind_page));
            if(!tp){
                fprintf(stderr, "malloc: error\n");
                exit(1);
            }
            tp->addr = addr >> 8;
            memset(tp->data, 0xff, 256);
            tp->next = rp;
            rp = tp;
        }else{
            while(tp->addr != addr >> 8){
                if(tp->next && tp->next->addr <= addr >> 8){
                    tp = tp->next;
                }else{
                    struct chaind_page *np;
                    np = malloc(sizeof(struct chaind_page));
                    if(!np){
                        fprintf(stderr, "malloc: error\n");
                        exit(1);
                    }
                    np->addr = addr >> 8;
                    memset(np->data, 0xff, 256);
                    np->next = tp->next;
                    tp->next = np;
                    tp = np;
                }
            }
        }
        for(int i = 0; i < bc-al-1; i++){
            if((addr + i) >> 8 != tp->addr){
                if(tp->next && tp->next->addr <= addr >> 8){
                    tp = tp->next;
                }else{
                    struct chaind_page *np;
                    np = malloc(sizeof(struct chaind_page));
                    if(!np){
                        fprintf(stderr, "malloc: error\n");
                        exit(1);
                    }
                    np->addr = (addr + i) >> 8;
                    memset(np->data, 0xff, 256);
                    np->next = tp->next;
                    tp->next = np;
                    tp = np;
                }
            }
            tp->data[(addr+i)&0xff] = 0;
            for(int j = 4+(al+i)*2; j < 4+(al+i)*2+2; j++){
                tp->data[(addr+i)&0xff] <<= 4;
                if('0' <= linebuf[j] && linebuf[j] <= '9'){
                    tp->data[(addr+i)&0xff] |= linebuf[j] - '0';
                }else if('A' <= linebuf[j] && linebuf[j] <= 'F'){
                    tp->data[(addr+i)&0xff] |= linebuf[j] - 'A' + 10;
                }else{
                    free_cpage(rp);
                    return NULL;
                }
            }
            sum += tp->data[(addr+i)&0xff];
        }

        uint8_t wsum = 0;
        for(int i = 4+bc*2-2; i < 4+bc*2; i++){
            wsum <<= 4;
            if('0' <= linebuf[i] && linebuf[i] <= '9'){
                wsum |= linebuf[i] - '0';
            }else if('A' <= linebuf[i] && linebuf[i] <= 'F'){
                wsum |= linebuf[i] - 'A' + 10;
            }else{
                free_cpage(rp);
                return NULL;
            }
        }
        if(wsum != (uint8_t)~sum){
            free_cpage(rp);
            return NULL;
        }
    }
    return rp;
};

char *get_config_file(void){
#ifdef R8C_WRITER_CONFIG_AT
    for(int i = 0; i < sizeof(path_to_conf)/sizeof(path_to_conf[0]); i++){
#ifdef _WIN32
        DWORD rv = GetFileAttributes(path_to_conf[i]);
        if(rv != -1 && !(rv & FILE_ATTRIBUTE_DIRECTORY)){
            return path_to_conf[i];
        }
#else
        struct stat fs;
        if(stat(path_to_conf[i], &fs)){
            continue;
        }else if(S_ISREG(fs.st_mode)){
            return path_to_conf[i];
        }
#endif
    }
#endif
    return NULL;
}

uint32_t read_duint_seg(FILE *fh){
    uint32_t ret = 0;
    int c;
    while((c = getc(fh)) != EOF){
        if('0' <= c && c <= '9'){
            ret *= 10;
            ret += c - '0';
        }else{
            ungetc(c, fh);
            break;
        }
    }
    return ret;
}

uint32_t read_huint_seg(FILE *fh){
    uint32_t ret = 0;
    int c;
    while((c = getc(fh)) != EOF){
        if('0' <= c && c <= '9'){
            ret *= 16;
            ret += c - '0';
        }else if('A' <= c && c <= 'F'){
            ret *= 16;
            ret += c - 'A' + 10;
        }else if('a' <= c && c <= 'f'){
            ret *= 16;
            ret += c - 'a' + 10;
        }else{
            ungetc(c, fh);
            break;
        }
    }
    return ret;
}

void read_str_seg(FILE *fh, struct vl_string *dst){
    int c;
    if(getc(fh) != '"'){
        fprintf(stderr, "\" needed before string.\n");
        exit(1);
    }
    while((c = getc(fh)) != EOF){
        if(c == '"'){
            break;
        }else if(c == '\\'){
            vl_string_push(dst, getc(fh));
        }else{
            vl_string_push(dst, c);
        }
    }
    if(c != '"'){
        fprintf(stderr, "\" needed after string.\n");
        exit(1);
    }
    vl_string_push(dst, 0);
}

void load_config(char *cfg_file){
    if(!cfg_file){
        fprintf(stderr, "config file not found.\n");
        exit(1);
    }

    FILE *fs = fopen(cfg_file, "r");
    if(!fs){
        perror("load_config:fopen");
    }

    struct writer_desc_chain *writer_tail = NULL;
    struct target_mmap_chain *target_tail = NULL;

    int section = 0;
    int subsection = 0;
    int nest = 0;
    struct vl_string string;
    vl_string_new(&string, 1024);
    int aldat = 0;
    int aldats = 0;
    int c;
    while((c = getc(fs)) != EOF){
        if(c == '#'){
            while(1){
                c = getc(fs);
                if(c == '\n' || c == EOF){
                    break;
                }
            }
        }else if(c == '{'){
            if(nest == 0 && !section){
                fprintf(stderr, "some errors in config file\n");
                exit(1);
            }
            nest++;
            if(nest > 2){
                fprintf(stderr, "some errors in config file\n");
                exit(1);
            }

            if(nest == 1){
                subsection = 0;
                if(section == 1){       //writer
                    struct writer_desc_chain *nw
                            = malloc(sizeof(struct writer_desc_chain));
                    if(!nw){
                        fprintf(stderr, "malloc: error\n");
                        exit(1);
                    }
                    memset(nw, 0, sizeof(*nw));
                    nw->config.timeout = 1000;
                    if(!writers){
                        assert(writer_tail == NULL);
                        writers = writer_tail = nw;
                    }else{
                        writer_tail->next = nw;
                        writer_tail = nw;
                    }
                }else{                  //target
                    struct target_mmap_chain *nt
                            = malloc(sizeof(struct target_mmap_chain));
                    if(!nt){
                        fprintf(stderr, "malloc: error\n");
                        exit(1);
                    }
                    memset(nt, 0, sizeof(*nt));
                    if(!targets){
                        assert(target_tail == NULL);
                        targets = target_tail = nt;
                    }else{
                        target_tail->next = nt;
                        target_tail = nt;
                    }
                }
                aldat = 0;
            }else if(nest == 2){
                aldats = 0;
            }
        }else if(c == '}'){
            nest--;
            if(nest < 0){
                fprintf(stderr, "some errors in config file\n");
                exit(1);
            }
            if(nest == 0){
                if(section == 1){       //writer
                    if(!writer_tail->name){
                        fprintf(stderr, "some errors in config file\n");
                        exit(1);
                    }
                }else{                  //target
                    if(!target_tail->friendry_name){
                        fprintf(stderr, "some errors in config file\n");
                        exit(1);
                    }
                }
            }else if(nest == 1){
                if(section == 1){
                    fprintf(stderr, "some errors in config file\n");
                    exit(1);
                }else{
                    if(subsection == 1){
                        assert(target_tail->namen == 0);
                        assert(string.buf[string.length-1] == 0);
                        for(size_t i = 0; i < string.length; i++){
                            if(!string.buf[i]){
                                target_tail->namen++;
                            }
                        }
                        target_tail->names
                            = malloc(sizeof(char *) * target_tail->namen);
                        char *sd = malloc(string.length);
                        if(!target_tail->names || !sd){
                            fprintf(stderr, "malloc: error\n");
                            exit(1);
                        }
                        target_tail->names[0] = sd;
                        size_t n = 1;
                        for(size_t i = 0; i < string.length; i++){
                            sd[i] = string.buf[i];
                            if(!string.buf[i] && n < target_tail->namen){
                                target_tail->names[n++] = &sd[i+1];
                            }
                        }
                        if(n != target_tail->namen){
                            fprintf(stderr, "some errors in config file\n");
                            exit(1);
                        }
                    }else if(subsection == 2){
                        assert(target_tail->blockn == 0);
                        if(string.length % 8){
                            fprintf(stderr, "some errors in config file\n");
                            exit(1);
                        }
                        target_tail->blockn = string.length / 8;
                        target_tail->blocks = malloc(sizeof(struct flash_block)
                                                         * target_tail->blockn);
                        uint32_t *sp = (uint32_t *)string.buf;
                        if(!target_tail->blocks){
                            fprintf(stderr, "malloc: error\n");
                            exit(1);
                        }
                        for(size_t i = 0; i < target_tail->blockn; i++){
                            target_tail->blocks[i].start = sp[i*2];
                            target_tail->blocks[i].size = sp[i*2+1];
                        }
                    }else{
                        fprintf(stderr, "some errors in config file\n");
                        exit(1);
                    }
                }
                string.length = 0;
            }
        }else if(c == ' ' || c == '\n'){
            //do nothing
        }else if(nest == 0){
            if(c == 'w'){
                section = 1;
            }else if(c == 't'){
                section = 2;
            }else{
                fprintf(stderr, "some errors in config file\n");
                exit(1);
            }
        }else if(nest == 1){
            if(c == ','){
                if(subsection++ < 0){
                    fprintf(stderr, "some errors in config file\n");
                    exit(1);
                }
                aldat = 0;
                continue;
            }else if(aldat){
                fprintf(stderr, "some errors in config file\n");
                exit(1);
            }
            aldat = 1;
            ungetc(c, fs);
            if(section == 1){       //writer
                if(subsection == 0){
                    read_str_seg(fs, &string);
                    assert(string.buf[string.length-1] == 0);
                    writer_tail->name = malloc(string.length);
                    if(!writer_tail->name){
                        fprintf(stderr, "malloc: error\n");
                        exit(1);
                    }
                    strcpy(writer_tail->name, string.buf);
                    string.length = 0;
                }else if(subsection == 1){
                    read_str_seg(fs, &string);
                    assert(string.buf[string.length-1] == 0);
                    writer_tail->default_path = malloc(string.length);
                    if(!writer_tail->default_path){
                        fprintf(stderr, "malloc: error\n");
                        exit(1);
                    }
                    strcpy(writer_tail->default_path, string.buf);
                    string.length = 0;
                }else if(subsection == 2){
                    writer_tail->config.pin_reset = read_huint_seg(fs);
                }else if(subsection == 3){
                    writer_tail->config.pin_mode = read_huint_seg(fs);
                }else if(subsection == 4){
                    writer_tail->config.flags = read_huint_seg(fs);
                }else if(subsection == 5){
                    writer_tail->config.timeout = read_duint_seg(fs);
                }else{
                    fprintf(stderr, "some errors in config file\n");
                    exit(1);
                }
            }else{                  //target
                if(subsection == 0){
                    read_str_seg(fs, &string);
                    assert(string.buf[string.length-1] == 0);
                    target_tail->friendry_name = malloc(string.length);
                    if(!target_tail->friendry_name){
                        fprintf(stderr, "malloc: error\n");
                        exit(1);
                    }
                    strcpy(target_tail->friendry_name, string.buf);
                    string.length = 0;
                }else{
                    fprintf(stderr, "some errors in config file\n");
                    exit(1);
                }
            }
        }else if(nest == 2){
            if(c == ','){
                aldats = 0;
                continue;
            }else if(aldats){
                fprintf(stderr, "some errors in config file\n");
                exit(1);
            }
            aldats = 1;
            if(section != 2){
                fprintf(stderr, "some errors in config file\n");
                exit(1);
            }

            ungetc(c, fs);
            if(subsection == 1){
                read_str_seg(fs, &string);
            }else if(subsection == 2){
                uint32_t val;
                if(string.length % 8){      //size
                    val = read_duint_seg(fs);
                    int u = getc(fs);
                    if(u == 'k' || u == 'K'){
                        val *= 1024;
                    }else{
                        ungetc(u, fs);
                    }
                }else{                      //address
                    val = read_huint_seg(fs);
                }
                for(int i = 0; i < 32; i+=8){
                    vl_string_push(&string, (val >> i) & 0xff);
                }
            }else{
                fprintf(stderr, "some errors in config file\n");
                exit(1);
            }
        }else{
            fprintf(stderr, "some errors in config file\n");
            exit(1);
        }
    }
    vl_string_free(&string);
    fclose(fs);
}

struct writer_desc_chain *search_writer(char *name){
    struct writer_desc_chain *wp = writers;
    while(wp){
        if(!strcmp(wp->name, name)){
            return wp;
        }
        wp = wp->next;
    }
    return NULL;
}

struct target_mmap_chain *search_target(char *name){
    char *sts = malloc(strlen(name) + 1);
    if(!sts){
        fprintf(stderr, "malloc: error\n");
        exit(1);
    }
    strcpy(sts, name);
    strctarget(sts);

    struct target_mmap_chain *mp = targets;
    while(mp){
        for(int i = 0; i < mp->namen; i++){
            if(!strcmp(mp->names[i], sts)){
                free(sts);
                return mp;
            }
        }
        mp = mp->next;
    }
    free(sts);
    return NULL;
}

void say_hello(void){
    printf("*** %s %s ***\n", myname, version);
    printf("%s\n", copyright);
    printf("%s\n\n", lisence);
}

void show_usage(char *argv0){
    size_t optn = sizeof(arg_def)/sizeof(arg_def[0]);

    printf("%s %s (%s)\n", myname, version, whoami);
    printf("%s\n\n", copyright);
    printf("usage: %s options filename\n\n", argv0);
    printf("options:\n");
    for(size_t i = 0; i < optn; i++){
        printf("  ");
        if(arg_def[i].shortname){
            printf("-%c", arg_def[i].shortname);
            if(arg_def[i].ifparam){
                printf(" param");
            }
            if(arg_def[i].longname){
                printf(", ");
            }
        }
        if(arg_def[i].longname){
            printf("--%s ", arg_def[i].longname);
            if(arg_def[i].ifparam){
                printf("param");
            }
        }
        printf("\n");
        printf("    %s\n", arg_def[i].description);
    }

    printf("\nfileformat:\n");
    printf("  Only Motorola S-record supported.\n");
}

void interpret_all_opts(int argc, char **argv){
    size_t optn = sizeof(arg_def)/sizeof(arg_def[0]);
    char *optstring = malloc(optn * 2 + 1);
    if(!optstring){
        fprintf(stderr, "malloc: error\n");
        exit(1);
    }
    char *osp = optstring;
#ifdef R8C_WRITER_LONGOPT
    struct option *longopts = malloc((optn + 1) * sizeof(struct option));
    if(!longopts){
        fprintf(stderr, "malloc: error\n");
        exit(1);
    }
    struct option *lop = longopts;
#endif

    int returnindex = 0;

    for(size_t i = 0; i < optn; i++){
        if(arg_def[i].shortname){
            *(osp++) = arg_def[i].shortname;
            if(arg_def[i].ifparam){
                *(osp++) = ':';
            }
        }
#ifdef R8C_WRITER_LONGOPT
        if(arg_def[i].longname){
            lop->name = arg_def[i].longname;
            lop->has_arg = arg_def[i].ifparam;
            lop->flag = &returnindex;
            lop->val = i;
            lop++;
        }
#endif
    }
    osp = '\0';
#ifdef R8C_WRITER_LONGOPT
    memset(lop, 0, sizeof(*lop));
#endif

    while(1){
#ifdef R8C_WRITER_LONGOPT
        int rv = getopt_long(argc, argv, optstring, longopts, NULL);
#else
        int rv = getopt(argc, argv, optstring);
#endif
        struct myoption *mip = NULL;
        if(rv == -1){
            break;
        }else if(rv == 0){
            mip = &arg_def[returnindex];
        }else{
            for(size_t i = 0; i < optn; i++){
                if(arg_def[i].shortname == rv){
                    mip = &arg_def[i];
                    break;
                }
            }
        }

        if(!mip){
            show_usage(argv[0]);
            exit(1);
        }

        if(mip->ifparam){
            size_t al = strlen(optarg);
            mip->param = malloc(al+1);
            if(!mip->param){
                fprintf(stderr, "malloc: error\n");
                exit(1);
            }
            strcpy(mip->param, optarg);
        }else{
            mip->param = malloc(1);
            if(!mip->param){
                fprintf(stderr, "malloc: error\n");
                exit(1);
            }
            *(char *)mip->param = 0;
        }
    }

    free(optstring);
#ifdef R8C_WRITER_LONGOPT
    free(longopts);
#endif

    if(argc != optind){
        if(argc != optind + 1){
            fprintf(stderr, "only one input file can be set\n");
            show_usage(argv[0]);
            exit(1);
        }
        filename = argv[optind];
    }
}

int main(int argc, char **argv){
    interpret_all_opts(argc, argv);

    if(arg_def[OPT_HELP].param){
        show_usage(argv[0]);
        return 0;
    }
    if(arg_def[OPT_VERSION].param){
        say_hello();
        return 0;
    }

    if(arg_def[OPT_CONFIG].param){
        load_config(arg_def[OPT_CONFIG].param);
    }else{
        char *cfg = get_config_file();
        if(!cfg){
            printf("No config file detected automatically.\n");
            exit(1);
        }
        load_config(cfg);
    }

    struct writer_desc_chain *writer_use = NULL;
    struct target_mmap_chain *target_use = NULL;
    if(arg_def[OPT_WRITER].param){
        writer_use = search_writer(arg_def[OPT_WRITER].param);
        if(!writer_use){
            printf("Undefined writer %s.\n", arg_def[OPT_WRITER].param);
            exit(1);
        }
    }
    if(arg_def[OPT_TARGET].param){
        target_use = search_target(arg_def[OPT_TARGET].param);
        if(!target_use){
            printf("Undefined target %s.\n", arg_def[OPT_TARGET].param);
            exit(1);
        }
    }

    if(arg_def[OPT_SHOWINFO].param){
        if(!writer_use && !target_use){
            printf("Give target and/or writer\n");
            exit(1);
        }
        say_hello();

        if(writer_use){
            printf("name: %s\n", writer_use->name);
            printf("path: %s\n", writer_use->default_path);
            printf("reset/mode: %2x/%2x\n", writer_use->config.pin_reset,
                                            writer_use->config.pin_mode);
            printf("flag/timeout: %2x/%d\n", writer_use->config.flags,
                                             writer_use->config.timeout);
        }

        if(target_use){
            printf("target: %s\n", target_use->friendry_name);
            printf("names:");
            for(int i = 0; i < target_use->namen; i++){
                printf(" %s", target_use->names[i]);
            }
            printf("\nflash memory blocks:\n");
            for(int i = 0; i < target_use->blockn; i++){
                printf("%06x: %dbytes\n",
                        target_use->blocks[i].start,
                        target_use->blocks[i].size);
            }
        }
        return 0;
    }

    if(!writer_use || !target_use){
        printf("Give target and writer\n");
        exit(1);
    }

    if(!filename && (!arg_def[OPT_TEST].param && !arg_def[OPT_EXEC].param)){
        printf("Specify the input file name\n");
        exit(1);
    }
    char *targ_dev;
    if(arg_def[OPT_DEVICE].param){
        targ_dev = arg_def[OPT_DEVICE].param;
    }else{
        targ_dev = writer_use->default_path;
        if(!targ_dev){
            printf("Default device not confugured.\n");
            exit(1);
        }
    }

    say_hello();

    int es;
    FILE *fp = NULL;
    struct chaind_page *cp = NULL;
    if(!arg_def[OPT_TEST].param && !arg_def[OPT_EXEC].param){
        fp = fopen(filename, "r");
        if(!fp){
            perror("fopen: failed to open the input file");
            exit(1);
        }
        cp = srecfp2cpage(fp);
        if(!cp){
            printf("srecord error\n");
            exit(1);
        }
        fclose(fp);
    }

    struct chaind_page *tcp = cp;
    char *used_blockmap = malloc(target_use->blockn);
    if(!used_blockmap){
        printf("malloc: error\n");
        exit(1);
    }
    memset(used_blockmap, 0, target_use->blockn);
    while(tcp){
        for(size_t i = 0; i < target_use->blockn; i++){
            if(target_use->blocks[i].start <= tcp->addr * 0x100u &&
               target_use->blocks[i].start + target_use->blocks[i].size
                                            > tcp->addr * 0x100u){
                used_blockmap[i] = 1;
            }
        }
        tcp = tcp->next;
    }

    struct writer_context wc;
    wc = writer_use->config;

#ifdef _WIN32
    wc.fh = CreateFile(targ_dev, GENERIC_READ | GENERIC_WRITE,
                       0, 0, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, 0);
    if(wc.fh == INVALID_HANDLE_VALUE){
        if(GetLastError() == ERROR_FILE_NOT_FOUND){
            printf("%s can't be opened: Not found\n", targ_dev);
        }else{
            printf("%s can't be opened\n", targ_dev);
        }
        exit(1);
    }
#else
    wc.fd = open(targ_dev, O_RDWR);
    if(wc.fd < 0){
        perror("open: device file of writer");
        exit(1);
    }
#endif

    if(arg_def[OPT_EXEC].param){
        serdev_init(&wc);

        set_mode(&wc, 1);
        set_reset(&wc, 0);

        my_msleep(35);
        set_reset(&wc, 1);       //reset pin to high
        printf("target reset done.\n");
        return 0;
    }

    printf("hand shaking...");
    if((es = program_init(&wc))){
        printf("\nfaild to start communication. code:%d\n", es);
        exit(1);
    }

    if((wc.status & WRSTAT_IDSTAT) == WRSTAT_NOIDCK){
        printf("No id ckeck needed.\n");
    }else{
        //                  ,   ,   ,
        uint64_t id = 0x00ffffffffffffff;
        if(arg_def[OPT_UNLOCKID].param){
            char *p = arg_def[OPT_UNLOCKID].param;
            id = 0;
            while(*p){
                if('0' <= *p && *p <= '9'){
                    id *= 0x10;
                    id += *p - '0';
                }else if('A' <= *p && *p <= 'F'){
                    id *= 0x10;
                    id += *p - 'A' + 0xA;
                }else if('a' <= *p && *p <= 'f'){
                    id *= 0x10;
                    id += *p - 'a' + 0xa;
                }
                p++;
            }
            //          ,   ,   ,
            id &= 0x00ffffffffffffff;
        }
        printf("sending id...");
        int rv = program_check_id(&wc, id);
        if(rv){
            printf("\nID ckeck error: %d\n", rv);
            exit(1);
        }
    }
    printf("ok!\n");

    char ver_string[10];
    if(program_get_version(&wc, ver_string)){
        printf("getting version failed\n");
        exit(1);
    }
    printf("target firmware version: %s\n", ver_string);

    if(arg_def[OPT_TEST].param){
        goto exit_normal;
    }

    printf("erasing blocks.\n");
    for(size_t i = 0; i < target_use->blockn; i++){
        if(used_blockmap[i]){
            printf("erasing blocks %04x...\n",
                                   target_use->blocks[i].start >> 8);
            if(program_erase_block(&wc,
                                   target_use->blocks[i].start >> 8)){
                printf("eraseing error\n");
                exit(1);
            }
        }
    }

    printf("writing pages.\n");
    char buf[256];
    while(cp){
        memcpy(buf, cp->data, 256);
        printf("writing pages %04x...\n", cp->addr);
        es = program_swrite_page(&wc, cp->addr, buf);
        if(es){
            printf("writing error of page %04x. code:%d\n", cp->addr, es);
            if(es >= 20){
                printf("data on page %04x:", cp->addr);
                pagedump(buf);
            }
            exit(1);
        }
        cp = cp->next;
    }

exit_normal:
#ifdef _WIN32
    if(!CloseHandle(wc.fh)){
        printf("Error while closing handle\n");
        exit(1);
    }
#else
    if(close(wc.fd) < 0){
        perror("close");
        exit(1);
    }
#endif

    printf("done.\n");

    return 0;
}
